/* A VDelta implementation for Subversion.
 *
 * (See comments in make_vdelta() for algorithm details.)
 *
 * ================================================================
 * Copyright (c) 2000 Collab.Net.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 * 
 * 3. The end-user documentation included with the redistribution, if
 * any, must include the following acknowlegement: "This product includes
 * software developed by Collab.Net (http://www.Collab.Net/)."
 * Alternately, this acknowlegement may appear in the software itself, if
 * and wherever such third-party acknowlegements normally appear.
 * 
 * 4. The hosted project names must not be used to endorse or promote
 * products derived from this software without prior written
 * permission. For written permission, please contact info@collab.net.
 * 
 * 5. Products derived from this software may not use the "Tigris" name
 * nor may "Tigris" appear in their names without prior written
 * permission of Collab.Net.
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL COLLAB.NET OR ITS CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 * GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
 * IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * ====================================================================
 * 
 * This software may consist of voluntary contributions made by many
 * individuals on behalf of Collab.Net.
 */



#include <assert.h>
#include <stdlib.h>
#include <apr_pools.h>
#include "svn_delta.h"
#include "delta_stream.h"


/* Hash table implementation
   This is a fixed-size hash table that accomodates hash
   collisions, but not key collisions. */

/* Hash value type. */
typedef apr_uint32_t vd_hash_t;

/* This is a magical hash key that marks free slots in the table */
static char const free_slot[] = "free";

/* Hash table slot: */
typedef struct vd_slot_t vd_slot_t;
struct vd_slot_t {
  char const* key;              /* The slot's key. If key == free_slot,
                                   this slot is unused. Otherwise key
                                   will point into the data stream. */
  union {
    struct {                    /* If key != free_slot: */
      vd_hash_t hash;           /*   the key's hash value; */
      vd_slot_t* next;          /*   next slot in the sibling chain. */
    } data;
    struct {                    /* If key == free_slot: */
      vd_slot_t* prev;          /*   previous and next slots in the */
      vd_slot_t* next;          /*   free-slot DL list. */
    } free;
  } info;
};

/* Hash table. */
typedef struct vd_table_t {
  size_t count;                 /* Number of elements in the table. */
  size_t capacity;              /* Table capacity. */
  vd_slot_t data[1];            /* Slot array. data[0] is the head of
                                   the free-slot DL list. */
} vd_table_t;



/* Create a hash table with `capacity' slots, allocated from `pool'.
   Thread all slots into a doubly-linked list of free slots. */
static vd_table_t*
vd_create_table (size_t capacity, apr_pool_t* pool)
{
  vd_table_t* table =
    apr_palloc (pool, (sizeof (*table) + capacity * sizeof (table->data[0])));
  if (table)
    {
      size_t i;
      table->count = 0;
      table->capacity = capacity;
      for (i = 0; i <= capacity; ++i)
        {
          vd_slot_t* const slot = &table->data[i];
          slot->key = free_slot;
          slot->info.free.prev = slot - 1;
          slot->info.free.next = slot + 1;
        }

      table->data[capacity].info.free.next = &table->data[0];
      table->data[0].info.free.prev = &table->data[capacity];
    }
  return table;
}



#define VD_KEY_LEN 4            /* Length of key -- four chars from the
                                   data stream. */

/* Compare two keys */
static APR_INLINE svn_boolean_t
equal_keys (char const* k1, char const* k2)
{
  return (0 == memcmp (k1, k2, VD_KEY_LEN));
}

/* Hash function. We use a 2-universla multiplicative hash. */
static APR_INLINE vd_hash_t
key_to_hash (char const* key)
{
  unsigned char const* k = key;
  vd_hash_t hash = 0;
  int i;
  for (i = 0; i < VD_KEY_LEN; ++i)
    hash = hash * 97 + *k++ + 41;
  return hash;
}

/* Convert a hash into a hash slot pointer. Notice that
   dada[0] is skipped. */
static APR_INLINE vd_slot_t*
hash_to_slot (vd_table_t* table, vd_hash_t hash)
{
  return &table->data[1 + hash % table->capacity];
}



/* Search the hash table and find the best slot for the key. */
static enum
{
  vd_found,                     /* Key is in table */
  vd_alias_found,               /* Key is not in table, but alias is */
  vd_not_found                  /* Neither key nor alias are in table */
}
find_slot (vd_table_t* table, char const* key, vd_hash_t hash,
           vd_slot_t** r_key_slot,  /* Key's slot */
           vd_slot_t** r_prev_slot) /* Previous in sibling list */
{
  vd_slot_t* key_slot = hash_to_slot (table, hash);
  vd_slot_t* prev_slot = NULL;

  /* If key's home slot is empty, or the key in the home slot isn't
     key's alias, key can't be in the table. */
  if (key_slot->key == free_slot)
    {
      *r_key_slot = key_slot;
      *r_prev_slot = NULL;
      return vd_not_found;      /* The key is not in the table. */
    }

  if (key_slot != hash_to_slot (table, key_slot->info.data.hash))
    {
      *r_key_slot = key_slot;
      *r_prev_slot = key_slot;  /* The home slot isn't empty. */
      return vd_not_found;      /* The key is still not in the table. */
    }

  /* The home slot contains an alias of key.
     Scan the sibling list to find the key slot. */
  while (hash != key_slot->info.data.hash
         || !equal_keys (key, key_slot->key))
    {
      if (key_slot->info.data.next == NULL)
        {
          /* We didn't find the key ... */
          *r_key_slot = NULL;
          *r_prev_slot = key_slot;
          return vd_alias_found; /* ... but we know where to put it. */
        }

      prev_slot = key_slot;
      key_slot = key_slot->info.data.next;

      assert (key_slot->key != free_slot);
    }

  /* Now we've found it. */
  *r_key_slot = key_slot;
  *r_prev_slot = prev_slot;
  return vd_found;
}



/* Remove a slot from the free-slot DL list. */
static APR_INLINE void
remove_slot_from_freelist (vd_table_t* table, vd_slot_t* slot)
{
  assert (slot->key == free_slot);

  /* Remove the slot from the free list. */
  slot->info.free.prev->info.free.next = slot->info.free.next;
  slot->info.free.next->info.free.prev = slot->info.free.prev;

  /* Mark the slot used. */
  slot->key = NULL;
  slot->info.data.next = NULL;

  /* Update the used count. */
  ++table->count;
}

/* Remove a slot into the free-slot DL list. */
static APR_INLINE void
insert_slot_into_freelist (vd_table_t* table, vd_slot_t* slot)
{
  vd_slot_t* const root_slot = &table->data[0];
  assert (slot->key != free_slot);

  /* Append the slot to the free list. */
  slot->info.free.prev = root_slot->info.free.prev;
  slot->info.free.next = root_slot;
  root_slot->info.free.prev = slot;
  slot->info.free.prev->info.free.next = slot;

  /* Mark slot unused. */
  slot->key = free_slot;

  /* Update the used count. */
  --table->count;
}



/* Insert a key into the table. If `replace_old' is TRUE, an existing
   equal key will be replaced with the new one, otherwise the old
   key will remain in the table. */
static svn_boolean_t
vd_table_insert (vd_table_t* table, char const* key,
                 svn_boolean_t replace_old)
{
  vd_hash_t const hash = key_to_hash (key);
  vd_slot_t* key_slot;
  vd_slot_t* prev_slot;

  assert (key != free_slot);

  if (table->count == table->capacity)
    return FALSE;               /* The table is full */

  switch (find_slot (table, key, hash, &key_slot, &prev_slot))
    {
    case vd_found:
      /* The key is already in the table. */
      if (replace_old)
        {
          key_slot->key = key;
          return TRUE;
        }
      else
        return FALSE;

    case vd_not_found:
      /* The key is not in table, and it's the first key with this hash. */
      if (!prev_slot)
        {
          /* We're on a free slot. */
          remove_slot_from_freelist (table, key_slot);
        }
      else
        {
          /* The home slot is used by a key that doesn't belong there.
             Move it, and put our key there instead. */

          vd_slot_t* const new_slot = table->data[0].info.free.next;
          remove_slot_from_freelist (table, new_slot);

          /* The key in the home slot must have at least one alias,
             so walk the sibling list from its home slot. */

          prev_slot = hash_to_slot (table, key_slot->info.data.hash);
          assert (prev_slot != key_slot);

          while (prev_slot->info.data.next != key_slot)
            prev_slot = prev_slot->info.data.next;

          /* Move the offending key. */
          prev_slot->info.data.next = new_slot;
          *new_slot = *key_slot;
        }
      break;

    case vd_alias_found:
      /* *prev_slot is the key's predecessor. */
      {
        /* Find an empty slot for the key and stick it to
           the end of the sibling list. */
        key_slot = table->data[0].info.free.next;
        remove_slot_from_freelist (table, key_slot);
        prev_slot->info.data.next = key_slot;
      }
      break;

    default:
      /* Something's very wrong with the code ... */
      assert (!"find_slot returned a weird value");
      return FALSE;
    }

    /* Insert the key in the table */
    key_slot->key = key;
    key_slot->info.data.hash = hash;
    key_slot->info.data.next = NULL;

    /* That's it. */
    return TRUE;
}



/* Delete a key from the table (we're currently not using this). */
static svn_boolean_t
vd_table_delete (vd_table_t* table, char const* key)
{
  vd_hash_t const hash = key_to_hash (key);
  vd_slot_t* key_slot;
  vd_slot_t* prev_slot;

  assert (key != free_slot);

  if (table->count == 0)
    return FALSE;               /* The table is empty */

  switch (find_slot (table, key, hash, &key_slot, &prev_slot))
    {
    case vd_found:
      if (prev_slot)
        {
          /* Remove this node from the sibling list. */
          prev_slot->info.data.next = key_slot->info.data.next;
        }
      else if (key_slot->info.data.next != NULL)
          {
            /* Move the next node into the home slot */
            prev_slot = key_slot;
            key_slot = key_slot->info.data.next;
            *prev_slot = *key_slot;
          }

      /* Free the deleted slot */
      insert_slot_into_freelist (table, key_slot);
      return TRUE;

    case vd_not_found:
    case vd_alias_found:
      return FALSE;

    default:
      /* Something's very wrong with the code ... */
      assert (!"find_slot returned a weird value");
      return FALSE;
    }
}



/* Find a key in the table. */
static char const*
vd_table_find (vd_table_t* table, void const* key)
{
  vd_hash_t const hash = key_to_hash (key);
  vd_slot_t* key_slot;
  vd_slot_t* prev_slot;

  assert (key != free_slot);

  if (table->count == 0)
    return NULL;                /* The table is empty */

  switch (find_slot (table, key, hash, &key_slot, &prev_slot))
    {
    case vd_found:
      return key_slot->key;

    case vd_not_found:
    case vd_alias_found:
      return NULL;

    default:
      /* Something's very wrong with the code ... */
      assert (!"find_slot returned a weird value");
      return NULL;
    }
}



static svn_boolean_t vd_replace_old = TRUE;

/* Core vdelta generator. */
static svn_error_t*
do_vdelta (svn_delta_window_t* window,
           char const* const buffer,
           apr_off_t source_len,
           apr_off_t target_len,
           apr_pool_t* pool)
{
  svn_error_t* err = NULL;
  char const* const begin = buffer;
  char const* const target = begin + source_len;
  char const* const end = target + target_len;
  char const* insert_from = NULL;
  char const* here = begin;

  /* Allocate the table with as many slots as there are source and
     target bytes. This will often be too many, but will at least
     never be too few. We can optimize for space some other time. */
  vd_table_t* table = vd_create_table (source_len + target_len, pool);

  /* Now, let's rock. See the description of the vdelta algorithm
     in "Delta Algorithms: An Empirical Analysis" by Hunt, Vo and Tichy. */
  for (;;)
    {
      char const* key = vd_table_find (table, here);
      if (key == NULL)
        {
          vd_table_insert (table, here, FALSE);
          if (here >= target && insert_from == NULL)
            {
              if (insert_from == NULL)
                insert_from = here;
            }
          ++here;
        }
      else
        {
          char const* match_start = key;
          char const* match_from = here;

          /* Commit the last insert  */
          if (here >= target && insert_from != NULL)
            err = svn_tdelta__insert_op (window, svn_delta_new,
                                         0, here - insert_from, insert_from);
          insert_from = NULL;
          if (err)
            break;

          /* Find the match length and insert the copy. */
          while (here < end && *here == *key)
            {
              ++here;
              ++key;
            }
          if (here >= target)
            {
              /* Adjust the destination of the copy if it straddles
                 the boundary. */
              if (match_from < target)
                {
                  match_start += target - match_from;
                  match_from = target;
                }

              /* Generate the copy.  If the source of the copy straddles
                 the boundary, we need to generate two copies, because
                 we can't represent a mixed copy in the window. */
              if (match_start < target && key > target)
                {
                  err = svn_tdelta__insert_op (window, svn_delta_source,
                                               match_start - begin,
                                               target - match_start,
                                               NULL);
                  if (!err)
                    err = svn_tdelta__insert_op (window, svn_delta_target,
                                                 0 /* Yes, that's right! */,
                                                 key - target,
                                                 NULL);
                }
              else if (key <= target) /* Copy from source. */
                err = svn_tdelta__insert_op (window, svn_delta_source,
                                             match_start - begin,
                                             here - match_from,
                                             NULL);
              else              /* Copy from target */
                err = svn_tdelta__insert_op (window, svn_delta_target,
                                             match_start - target,
                                             here - match_from,
                                             NULL);

              if (err)
                break;

              /* Insert the last indexes of the match into the table */
              if (end - here >= VD_KEY_LEN)
                {
                  char const* last;
                  for (last = here - VD_KEY_LEN + 1; last < here; ++last)
                    vd_table_insert (table, last, vd_replace_old);
                }
            }
        }

      /* If we're near the end, insert whatever's left. */
      if (end - here < VD_KEY_LEN)
        {
          if (insert_from != NULL)
            err = svn_tdelta__insert_op (window, svn_delta_new,
                                         0, end - insert_from, insert_from);
          else if (here < end)
            err = svn_tdelta__insert_op (window, svn_delta_new,
                                         0, end - here, here);
          break;
        }
    }

  return err;
}



/* Create a vdelta. */
svn_error_t*
svn_tdelta__vdelta (svn_delta_window_t** window,
                    svn_delta_stream_t* stream)
{
  svn_error_t* err = NULL;
  apr_off_t source_len = svn_tdelta__window_size;
  apr_off_t target_len = svn_tdelta__window_size;
  apr_pool_t* const pool = apr_make_sub_pool (stream->pool, NULL);
  char* const buffer = apr_palloc (pool, source_len + target_len);

  /* Read the source and target streams. */
  stream->source_fn (stream->source_data, buffer, &source_len);
  stream->target_fn (stream->target_data, buffer + source_len, &target_len);

  /* Forget everything if there's no target data. */
  *window = NULL;
  if (target_len == 0)
    {
      stream->more = FALSE;
      apr_destroy_pool (pool);
      return NULL;
    }

  /* Create the window */
  err = svn_tdelta__init_window (window, stream);
  if (err)
    {
      apr_destroy_pool (pool);
      return err;
    }

  /* If the target data is small enough (for some definition of small
     enough), just insert it into the window and be done with it. */
  if (target_len < VD_KEY_LEN)
    {
      stream->more = FALSE;
      err = svn_tdelta__insert_op (*window, svn_delta_new,
                                   0, target_len, buffer + source_len);
    }
  else
    {
      err = do_vdelta (*window, buffer, source_len, target_len, pool);
    }

  /* That's it. */
  apr_destroy_pool (pool);
  if (err)
    {
      svn_free_delta_window (*window);
      *window = NULL;
    }
  return err;
}



/* 
 * local variables:
 * eval: (load-file "../svn-dev.el")
 * end:
 */
